@@ -2,8 +2,8 @@
 //! including primitive integers and `num-bigint`. The latter option is enabled optionally.
 //!
 //! To achieve fast modular arithmetics, convert integers to any [ModularInteger] implementation
-//! use static `new()` or associated [ModularInteger::new()]. [MontgomeryInt] is a builtin implementation
-//! of [ModularInteger] based on the Montgomery form.
+//! use static `new()` or associated [ModularInteger::convert()] functions. Some builtin implementations
+//! of [ModularInteger] includes [MontgomeryInt] and [MersenneInt].
 //!
 //! Example code:
 //! ```rust
@@ -15,7 +15,7 @@
 //!
 //! // convert integers into ModularInteger
 //! let mx = MontgomeryInt::new(x, m);
-//! let my = mx.new(y); // faster than static new()
+//! let my = mx.convert(y); // faster than static MontgomeryInt::new(y, m)
 //! assert_eq!((mx * my).residue(), x * y % m);
 //! ```
 //!
@@ -29,49 +29,58 @@ extern crate std;
 
 use core::ops::{Add, Mul, Neg, Sub};
 
-/// This trait describes core modular arithmetic operations
+/// Core modular arithmetic operations.
+///
+/// Note that all functions will panic if the modulus is zero.
 pub trait ModularCoreOps<Rhs = Self, Modulus = Self> {
     type Output;
 
     /// Return (self + rhs) % m
     fn addm(self, rhs: Rhs, m: Modulus) -> Self::Output;
 
-    /// Return (self + rhs) % m
+    /// Return (self - rhs) % m
     fn subm(self, rhs: Rhs, m: Modulus) -> Self::Output;
 
     /// Return (self * rhs) % m
     fn mulm(self, rhs: Rhs, m: Modulus) -> Self::Output;
+}
+
+/// Core unary modular arithmetics
+///
+/// Note that all functions will panic if the modulus is zero.
+pub trait ModularUnaryOps<Modulus = Self> {
+    type Output;
 
     /// Return (-self) % m and make sure the result is normalized in range [0,m)
     fn negm(self, m: Modulus) -> Self::Output;
-}
-
-// TODO (v0.3): split negm, invm and sqrtm to ModularUnaryOps
-// and split powm and logm to ModularPow
-
-// TODO (v0.3): checked_jacobi, checked_kronecker, checked_legendre
-/// This trait describes modular arithmetic operations
-pub trait ModularOps<Rhs = Self, Modulus = Self>: ModularCoreOps<Rhs, Modulus> {
-    /// Return (self ^ exp) % m
-    fn powm(self, exp: Rhs, m: Modulus) -> Self::Output;
 
     /// Calculate modular inverse (x such that self*x = 1 mod m).
     ///
     /// This operation is only available for integer that is coprime to `m`. If not,
     /// the result will be [None].
-    fn invm(self, m: Modulus) -> Option<Self::Output>
-    where
-        Self: Sized;
+    fn invm(self, m: Modulus) -> Option<Self::Output>;
 
-    /// Calculate Jacobi Symbol (a|n), where a is self
-    ///
-    /// # Panics
-    /// if n is negative or even
-    fn jacobi(self, n: Modulus) -> i8;
+    /// Calculate modular double ( x+x mod m)
+    fn dblm(self, m: Modulus) -> Self::Output;
 
-    /// Calculate Kronecker Symbol (a|n), where a is self
-    fn kronecker(self, n: Modulus) -> i8;
+    /// Calculate modular square ( x*x mod m )
+    fn sqm(self, m: Modulus) -> Self::Output;
 
+    // TODO: Modular sqrt aka Quadratic residue, follow the behavior of FLINT `n_sqrtmod`
+    // fn sqrtm(self, m: Modulus) -> Option<Self::Output>;
+    // REF: https://stackoverflow.com/questions/6752374/cube-root-modulo-p-how-do-i-do-this
+}
+
+/// Modular power functions
+pub trait ModularPow<Exp = Self, Modulus = Self> {
+    type Output;
+
+    /// Return (self ^ exp) % m
+    fn powm(self, exp: Exp, m: Modulus) -> Self::Output;
+}
+
+/// Math symbols related to modular arithmetics
+pub trait ModularSymbols<Modulus = Self> {
     /// Calculate Legendre Symbol (a|n), where a is self.
     ///
     /// Note that this function doesn't perform primality check, since
@@ -79,15 +88,54 @@ pub trait ModularOps<Rhs = Self, Modulus = Self>: ModularCoreOps<Rhs, Modulus> {
     ///
     /// # Panics
     /// if n is not prime
-    fn legendre(self, n: Modulus) -> i8;
+    #[inline]
+    fn legendre(&self, n: Modulus) -> i8 {
+        self.checked_legendre(n).expect("n shoud be a prime")
+    }
 
-    // TODO: Modular sqrt aka Quadratic residue, follow the behavior of FLINT `n_sqrtmod`
-    // fn sqrtm(self, m: Modulus);
+    /// Checked version of [legendre()][ModularSymbols::legendre], return [None] if n is not prime
+    fn checked_legendre(&self, n: Modulus) -> Option<i8>;
 
-    // TODO: Discrete log aka index, follow the behavior of FLINT `n_discrete_log_bsgs`
-    // fn logm(self, base: Modulus, m: Modulus);
+    /// Calculate Jacobi Symbol (a|n), where a is self
+    ///
+    /// # Panics
+    /// if n is negative or even
+    #[inline]
+    fn jacobi(&self, n: Modulus) -> i8 {
+        self.checked_jacobi(n)
+            .expect("the Jacobi symbol is only defined for non-negative odd integers")
+    }
+
+    /// Checked version of [jacobi()][ModularSymbols::jacobi], return [None] if n is negative or even
+    fn checked_jacobi(&self, n: Modulus) -> Option<i8>;
+
+    /// Calculate Kronecker Symbol (a|n), where a is self
+    fn kronecker(&self, n: Modulus) -> i8;
 }
 
+// TODO: Discrete log aka index, follow the behavior of FLINT `n_discrete_log_bsgs`
+// fn logm(self, base: Modulus, m: Modulus);
+
+/// Collection of common modular arithmetic operations
+pub trait ModularOps<Rhs = Self, Modulus = Self, Output = Self>:
+    ModularCoreOps<Rhs, Modulus, Output = Output>
+    + ModularUnaryOps<Modulus, Output = Output>
+    + ModularPow<Rhs, Modulus, Output = Output>
+    + ModularSymbols<Modulus>
+{
+}
+impl<T, Rhs, Modulus> ModularOps<Rhs, Modulus> for T where
+    T: ModularCoreOps<Rhs, Modulus, Output = T>
+        + ModularUnaryOps<Modulus, Output = T>
+        + ModularPow<Rhs, Modulus, Output = T>
+        + ModularSymbols<Modulus>
+{
+}
+
+/// Collection of operations similar to [ModularOps], but takes operands with references
+pub trait ModularRefOps: for<'r> ModularOps<&'r Self, &'r Self> + Sized {}
+impl<T> ModularRefOps for T where T: for<'r> ModularOps<&'r T, &'r T> {}
+
 /// Provides a utility function to convert signed integers into unsigned modular form
 pub trait ModularAbs<Modulus> {
     /// Return |self| % m
@@ -119,24 +167,26 @@ pub trait ModularInteger:
     ///
     /// This method should be perferred over the static
     /// constructor to prevent unnecessary overhead of pre-computation.
-    fn new(&self, n: Self::Base) -> Self;
+    fn convert(&self, n: Self::Base) -> Self;
+
+    // Calculate the value of self + self
+    fn double(self) -> Self;
+
+    // Calculate the value of self * self
+    fn square(self) -> Self;
 }
 
+// XXX: implement this trait for ff::PrimeField?
+
 mod barret;
 mod double;
+mod mersenne;
 mod monty;
 mod prim;
-mod mersenne;
 
 pub use double::{udouble, umax};
-#[cfg(std)]
-pub use monty::MontgomeryBigint;
-pub use monty::{Montgomery, MontgomeryInt};
 pub use mersenne::MersenneInt;
+pub use monty::{Montgomery, MontgomeryInt};
 
 #[cfg(feature = "num-bigint")]
 mod bigint;
-
-// tests for ModularOps goes here
-#[cfg(test)]
-mod tests;
